/***********************************************************************
*                                                                      *
*     Program to calculate the first kind modified Bessel function     *
*  of integer order N, for any REAL X, using the function BESSI(N,X).  *
*                                                                      *
* -------------------------------------------------------------------- *
*    SAMPLE RUN:                                                       *
*                                                                      *
*    (Calculate Bessel function for N=2, X=0.75).                      *
*                                                                      *
*    Bessel function of order 2 for X =  0.7500:                       *
*                                                                      *
*         Y = 0.073667                                                 *
*                                                                      *
* -------------------------------------------------------------------- *
*    Reference: From Numath Library By Tuan Dang Trong in Fortran 77.  *
*                                                                      *
*                               C++ Release 1.1 By J-P Moreau, Paris.  *
*                                        (www.jpmoreau.fr)             *
*                               Code size optimization by Ralph Schenn *
*                                                                      * 
*    Version 1.1: corected value of P4 in BESSIO (P4=1.2067492 and not *
*                 1.2067429) Aug. 2011.                                *
***********************************************************************/

// http://www.johndcook.com/blog/bessel_functions/

#include <tiams.h>
#include "Bessel.h"
#include "Debug.h"
#include "Stack.h"

void BessWrap(double (*funcPtr)(int, double))
{       
	Access_AMS_Global_Variables; 
	unsigned short ord;
	double arg;
        
	EStackIndex argptr = top_estack;

	// check number of parameters
	CheckNumArgs (2);        
        
	// get order of bessel function
	ord = estack_to_short_simp (argptr);
	
	// get argument of bessel function
	argptr = next_expression_index (argptr);                

	if (is_list (argptr)) {
		argptr--;
		
		push_quantum (END_TAG);
		
		while (ESTACK (argptr) != END_TAG) {
			arg = estack_number_to_Float_inf (argptr);
	    
			// push value of bessel function as result
			push_Float ((*funcPtr) (ord, arg));
		
			argptr = next_expression_index (argptr);                
		}

		push_quantum (LIST_TAG); 
		
		argptr = top_estack;
		
		// push reversed list
		push_reversed_tail (top_estack - 1); 
			
		// beginning of reversed list
		push_quantum (LIST_TAG);	
	
		// delete list in original order
		delete_expression (argptr);
	} else {
		arg = estack_number_to_Float_inf (argptr);
    
		// push value of bessel function as result
		push_Float ((*funcPtr) (ord, arg));
	} 
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

/* bessel function core calculation
used for unmodified and modified bessel function
used for bessel function of 0th and 1st order
array of coefficients
length of array of coefficients
argument of bessel function */
static double BesselCore(const double arr[], // ok
	unsigned short len, 
	double x)
{
	double s = arr[len - 2] + x * arr[len - 1];
	short i; 
	
	for (i = len - 3; i >= 0; i--) {
		s = arr[i] + x * s;
	}
	
	return s;
}

/* modified bessel function for first and second order
(the only major differnce between orders are the coefs) */
static double BessI01(char ord, // ok
	const double P[], 
	const double Q[], 
	double X)
{
	double Y, AX, BX;

	if (fabs(X) < 3.75) {
		Y = X / 3.75;
		Y *= Y;
		//AX = P[0] + Y * (P[1] + Y * (P[2] + Y * (P[3] + Y * (P[4] + Y * (P[5] + Y * P[6])))));
		AX = BesselCore (P, 7, Y);
		return ord == 0 ? AX : X * AX;
	} else {
		AX = fabs(X);
		Y = 3.75 / AX;
		BX = exp(AX) / sqrt(AX);
		//AX = Q[0] + Y * (Q[1] + Y * (Q[2] + Y * (Q[3] + Y * (Q[4] + Y * (Q[5] + Y * (Q[6] + Y * (Q[7] + Y * Q[8])))))));
		AX = BesselCore (Q, 9, Y);
		return AX * BX;
	}	
}

//  Auxiliary Bessel functions for N=0, N=1
static double BessI0(double X) // ok
{
	const double P[] = {1.0, 
		3.5156229, 
		3.0899424, 
		1.2067492,
		0.2659732, 
		0.360768e-1, 
		0.45813e-2};

	const double Q[] = {0.39894228, 
		0.1328592e-1,
		0.225319e-2,
		-0.157565e-2, 
		0.916281e-2, 
		-0.2057706e-1,
		0.2635537e-1, 
		-0.1647633e-1, 
		0.392377e-2};
		
	return BessI01(0, P, Q, X);		
}

static double BessI1(double X) // ok
{
	const double P[] = {0.5, 
		0.87890594, 
		0.51498869, 
		0.15084934,
		0.2658733e-1, 
		0.301532e-2, 
		0.32411e-3};

	const double Q[] = {0.39894228, 
		-0.3988024e-1, 
		-0.362018e-2,
		0.163801e-2, 
		-0.1031555e-1, 
		0.2282967e-1,
		-0.2895312e-1,
		0.1787654e-1, 
		-0.420059e-2};
      
	return BessI01(1, P, Q, X);		
}

/*----------------------------------------------------------------------
!     This subroutine calculates the first kind modified Bessel function
!     of integer order N, for any REAL X. We use here the classical
!     recursion formula, when X > N. For X < N, the Miller's algorithm
!     is used to avoid overflows. 
!     REFERENCE:
!     C.W.CLENSHAW, CHEBYSHEV SERIES FOR MATHEMATICAL FUNCTIONS,
!     MATHEMATICAL TABLES, VOL.5, 1962.
------------------------------------------------------------------------*/
double BessI(int N, double X) // ok
{
	const int IACC = 40; 
	float BIGNO = 1e10, BIGNI = 1e-10;
	float TOX, BIM, BI = 1, BIP = 0, BSI = 0;
	int J, M;
	
	N = abs (N);

	if (N == 0) { 
		return BessI0(X); 
	} else if (N == 1) { 
		return BessI1(X); 
	} else if (X == 0.0) { 
		return 0.0; 
	} else {
		TOX = 2.0 / X;
		M = (int) (2 * ((N + floor (sqrt (IACC * N)))));
	      
		for (J = M; J > 0; J--) {
			BIM = BIP + J * TOX * BI;
			BIP = BI;
			BI  = BIM;
	        
			if (fabs(BI) > BIGNO) {
				BI  *= BIGNI;
				BIP *= BIGNI;
				BSI *= BIGNI;
	        }
	        
			if (J == N) { 
				BSI = BIP; 
			}
		}
	      
		return BSI * BessI0 (X) / BI;
	}
}

static double Sign(double X, double Y) { // ok
    return Y < 0.0 ? -fabs(X) : fabs(X);
}

/* bessel function for first and second order
(the only major differnce between orders are the coefs) */
static double BessJ01(char ord,
    const double P[],
    const double Q[],
    const double R[],
    const double S[],
    const double C,
    double X)
{
    double AX, FR, FS, Z, FP, FQ, XX, Y, TMP;

    if (ord == 0 && X == 0.0) {
            return 1.0;
    }
    
    AX = fabs(X);
  
    if (AX < 8.0) {
        Y = X * X;
        //FR = R[0] + Y * (R[1] + Y * (R[2] + Y * (R[3] + Y * (R[4] + Y * R[5]))));
        FR = BesselCore (R, 6, Y);
        //FS = S[0] + Y * (S[1] + Y * (S[2] + Y * (S[3] + Y * (S[4] + Y * S[5]))));
        FS = BesselCore (S, 6, Y);
        TMP = FR / FS;
        
        if (ord == 1) {
            TMP *= X;
        }
    } else {
        Z = 8.0 / AX;
        Y = Z * Z;
        XX = AX - C;
        //FP = P[0] + Y * (P[1] + Y * (P[2] + Y * (P[3] + Y * P[4])));
        FP = BesselCore (P, 5, Y);
        //FQ = Q[0] + Y * (Q[1] + Y * (Q[2] + Y * (Q[3] + Y * Q[4])));
        FQ = BesselCore (Q, 5, Y);
        TMP = sqrt(P[5] / AX) * (cos(XX) * FP - Z * sin(XX) * FQ);
        
        if (ord == 1) {
            TMP *= Sign(S[5], X);
        }
    }
  
    return TMP;
}

/***********************************************************************
      This subroutine calculates the First Kind Bessel Function of
      order 0, for any real number X. The polynomial approximation by
      series of Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
      REFERENCES:
      M.ABRAMOWITZ,I.A.STEGUN, HANDBOOK OF MATHEMATICAL FUNCTIONS, 1965.
      C.W.CLENSHAW, NATIONAL PHYSICAL LABORATORY MATHEMATICAL TABLES,
      VOL.5, 1962.
************************************************************************/
static double BessJ0 (double X) { // ok
    const double P[] = {1.0, 
        -0.1098628627E-2, 
        0.2734510407E-4,
        -0.2073370639E-5, 
        0.2093887211E-6, 
        0.636619772};
                
    const double Q[] = {-0.1562499995E-1, 
        0.1430488765E-3, 
        -0.6911147651E-5,
        0.7621095161E-6, 
        -0.9349451520E-7};
            
    const double R[] = {57568490574.0, 
        -13362590354.0, 
        651619640.7,
        -11214424.18, 
        77392.33017, 
        -184.9052456};
            
    const double S[] = {57568490411.0, 
        1029532985.0, 
        9494680.718,
        59272.64853, 
        267.8532712, 
        1.0};
            
    return BessJ01(0, P, Q, R, S, 0.785398164, X);
}

/**********************************************************************
      This subroutine calculates the First Kind Bessel Function of
      order 1, for any real number X. The polynomial approximation by
      series of Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
      REFERENCES:
      M.ABRAMOWITZ,I.A.STEGUN, HANDBOOK OF MATHEMATICAL FUNCTIONS, 1965.
      C.W.CLENSHAW, NATIONAL PHYSICAL LABORATORY MATHEMATICAL TABLES,
      VOL.5, 1962.
***********************************************************************/
static double BessJ1 (double X) { // ok
    const double P[] = {1.0, 
        0.183105E-2, 
        -0.3516396496E-4, 
        0.2457520174E-5,
        -0.240337019E-6, 
        0.636619772};
            
    const double Q[] = {0.04687499995, 
        -0.2002690873E-3, 
        0.8449199096E-5,
        -0.88228987E-6, 
        0.105787412E-6};
            
    const double R[] = {72362614232.0, 
        -7895059235.0, 
        242396853.1,
        -2972611.439,
        15704.48260, 
        -30.16036606};
            
    const double S[] = {144725228442.0, 
        2300535178.0, 
        18583304.74,
        99447.43394, 
        376.9991397,  
        1.0};

    return BessJ01(1, P, Q, R, S, 2.35619491, X);
}

/************************************************************************
      This subroutine calculates the first kind modified Bessel function
      of integer order N, for any REAL X. We use here the classical
      recursion formula, when X > N. For X < N, the Miller's algorithm
      is used to avoid overflows.
      ----------------------------- 
      REFERENCE:
      C.W.CLENSHAW, CHEBYSHEV SERIES FOR MATHEMATICAL FUNCTIONS,
      MATHEMATICAL TABLES, VOL.5, 1962.
*************************************************************************/
double BessJ(int N, double X) { // ok
	const int IACC = 40; 
	const double BIGNO = 1e10, BIGNI = 1e-10;
    
	double TOX, BJM, BJ = 1, BJP = 0, SUM = 0, TMP = 0, R;
	int J, JSUM = 0, M;
	Boolean NEG = N < 0;

    N = abs(N);
    
	if (N == 0) {
		R = BessJ0(X);
	} else if (N == 1) {
		R = BessJ1(X);
	} else if (X == 0.0) {
		R = 0.0;
	} else {
		TOX = 2.0 / X;
	        
		if (X > N) {
			BJM = BessJ0(X);
			BJ  = BessJ1(X);
	        
			for (J = 1; J < N; J++) {
				BJP = J * TOX * BJ - BJM;
				BJM = BJ;
	        	BJ  = BJP;
			}
	        
	    	R = BJ;
	    } else {
			M = (int) (2 * ((N + floor(sqrt(1.0 * (IACC * N)))) / 2));
	        
			for (J = M; J > 0; J--) {
				BJM = J * TOX * BJ - BJP;
				BJP = BJ;
	            BJ  = BJM;
	          
				if (fabs(BJ) > BIGNO) {
					BJ  *= BIGNI;
					BJP *= BIGNI;
					TMP *= BIGNI;
					SUM *= BIGNI;
				}
	          
				if (JSUM != 0) {
					SUM += BJ;
				}
	          
				JSUM = 1 - JSUM;
	          
				if (J == N) {
					TMP = BJP;
				}
			}
	        
			SUM = 2.0 * SUM - BJ;
			R = TMP / SUM;
		}       
	}
	
	// when sign of order changes sign of function changes only for odd orders
	return NEG && N % 2 ? -R : R;
}

/* ---------------------------------------------------------------------
!     CALCUL DE LA FONCTION BESSEL MODIFIEE DU 3EME ESPECE D'ORDRE 0
!     POUR TOUT X REEL NON NUL.
!
!     CALCULATES THE THE MODIFIED BESSEL FUNCTION OF THE THIRD KIND OF 
!     ORDER ZERO FOR ANY POSITIVE REAL ARGUMENT, X.
! --------------------------------------------------------------------*/
static double BessK01(unsigned char ord,
	const double P[],
	const double Q[],
	double X)  
{
	Access_AMS_Global_Variables; 
	
	double Y, AX;

	if (X == 0.0) { 
		return FLOATTAB[FPI_POSINF];  //arbitrary big value
	}

	if (X <= 2.0) {	
		Y = X * X / 4.0;
		AX = (ord == 0 ? -1 : 1) * log (X / 2.0) * BessI (ord, X);		
		//AX + (P1 + Y * (P2 + Y * (P3 + Y * (P4 + Y * (P5 + Y * (P6 + Y * P7))))));  
		return AX + BesselCore (P, 7, Y);
	} else {
		Y = 2.0 / X;
		AX = exp (-X) / sqrt (X);
		//AX * (Q1 + Y * (Q2 + Y * (Q3 + Y * (Q4 + Y * (Q5 + Y * (Q6 + Y * Q7))))));   
		return AX * BesselCore (Q, 7, Y);
	}
}

/* ---------------------------------------------------------------------
!     CALCUL DE LA FONCTION BESSEL MODIFIEE DU 3EME ESPECE D'ORDRE 0
!     POUR TOUT X REEL NON NUL.
!
!     CALCULATES THE THE MODIFIED BESSEL FUNCTION OF THE THIRD KIND OF 
!     ORDER ZERO FOR ANY POSITIVE REAL ARGUMENT, X.
! --------------------------------------------------------------------*/
static double BessK0(double X)  
{
	const double P[] = {-0.57721566, 
		0.42278420, 
		0.23069756, 
		0.3488590e-1,
		0.262698e-2, 
		0.10750e-3, 
		0.74e-5
	};

	const double Q[] = {1.25331414, 
		-0.7832358e-1, 
		0.2189568e-1, 
		-0.1062446e-1,
		0.587872e-2, 
		-0.251540e-2, 
		0.53208e-3
	};

	return BessK01(0,
		P,
		Q,
		X);  
}

/* ------------------------------------------------------------------------
!     CALCUL DE LA FONCTION BESSEL MODIFIEE DE 3EME ESPECE D'ORDRE 1
!     POUR TOUT X REEL POSITF NON NUL.
!
!     CALCULATES THE THE MODIFIED BESSEL FUNCTION OF THE THIRD KIND OF 
!     ORDER ONE FOR ANY POSITIVE REAL ARGUMENT, X.
! -----------------------------------------------------------------------*/
static double BessK1(double X)  {
	const double P[] = {1.0, 
		0.15443144, 
		-0.67278579, 
		-0.18156897,
		-0.1919402e-1, 
		-0.110404e-2, 
		-0.4686e-4
	};

	const double Q[] = {1.25331414, 
		0.23498619, 
		-0.3655620e-1, 
		0.1504268e-1,
		-0.780353e-2, 
		0.325614e-2,
		-0.68245e-3
	};

	return BessK01(1,
		P,
		Q,
		X);  
}

/* -----------------------------------------------------------------------
!     CE SOUS-PROGRAMME CALCULE LA FONCTION BESSEL MODIFIFIEE 3E ESPECE
!     D'ORDRE N ENTIER POUR TOUT X REEL POSITIF > 0.  ON UTILISE ICI LA
!     FORMULE DE RECURRENCE CLASSIQUE EN PARTANT DE BESSK0 ET BESSK1.
!
!     THIS ROUTINE CALCULATES THE MODIFIED BESSEL FUNCTION OF THE THIRD
!     KIND OF INTEGER ORDER, N FOR ANY POSITIVE REAL ARGUMENT, X. THE
!     CLASSICAL RECURSION FORMULA IS USED, STARTING FROM BESSK0 AND BESSK1.
! ------------------------------------------------------------------------ 
!     REFERENCE:
!     C.W.CLENSHAW, CHEBYSHEV SERIES FOR MATHEMATICAL FUNCTIONS,
!     MATHEMATICAL TABLES, VOL.5, 1962.
! ----------------------------------------------------------------------*/
double BessK(int N, double X) 
{
	Access_AMS_Global_Variables; 

	double TOX, BK, BKM, BKP;
	int J;
	
	N = abs (N);

	if (N == 0) {
		return BessK0 (X);
	} else if (N == 1) { 
		return BessK1 (X);
	} else if (X == 0.0) { 
		return FLOATTAB[FPI_POSINF];  //arbitrary big value
	} else {
		TOX = 2.0 / X;
		BK  = BessK1 (X);
		BKM = BessK0 (X);
	
		for (J = 1; J < N; J++) {
			BKP = BKM + J * TOX * BK;
			BKM = BK;
			BK  = BKP;
		}
	
		return BK;
	}
}

/* --------------------------------------------------------------------
      This subroutine calculates the Second Kind Bessel Function of
      order 0, for any real number X. The polynomial approximation by
      series of Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
      REFERENCES:
      M.ABRAMOWITZ,I.A.STEGUN, HANDBOOK OF MATHEMATICAL FUNCTIONS, 1965.
      C.W.CLENSHAW, NATIONAL PHYSICAL LABORATORY MATHEMATICAL TABLES,
      VOL.5, 1962.
  --------------------------------------------------------------------- */
static double BessY01 (char ord,
	const double P[], 
	const double Q[],
	const double R[], 
	const double S[],
	double X) 
{
	Access_AMS_Global_Variables;

	double FS, FR, Z, FP, FQ, XX, Y;
      
	if (X == 0.0) {
		return FLOATTAB[FPI_NEGINF];
	}
 	  
	if (X < 8.0) {
		Y = X * X;
		//FR = R1 + Y * (R2 + Y * (R3 + Y * (R4 + Y * (R5 + Y * R6)))); 
		FR = BesselCore (R, 6, Y);
		//FS = S1 + Y * (S2 + Y * (S3 + Y * (S4 + Y * (S5 + Y * S6)))); 
		FS = BesselCore (S, (ord == 0 ? 6 : 7), Y);      
		
		// only the writers of Sierra C know why we can not put the log
		// directly into the returned expression (compiler error)
		XX = log (X);
		return (ord == 0 ? 1 : X) * (FR / FS) + 
			0.636619772 * (BessJ (ord, X) * XX - (ord == 0 ? 0 : 1.0 / X));
	} else {
		Z = 8.0 / X;
		Y = Z * Z;		
		XX = X - (ord == 0 ? 0.785398164 : 2.356194491);
		//FP = P1 + Y * (P2 + Y * (P3 + Y * (P4 + Y * P5)));
		FP = BesselCore (P, 5, Y);
		//FQ = Q1 + Y * (Q2 + Y * (Q3 + Y * (Q4 + Y * Q5)));
		FQ = BesselCore (Q, 5, Y);		
		// only the writers of Sierra C know why we can not put the log
		// directly into the returned expression (compiler error)
		X = sqrt (0.636619772 / X);
		return X * (FP * sin (XX) + Z * FQ * cos (XX));
	}
}

/* --------------------------------------------------------------------
      This subroutine calculates the Second Kind Bessel Function of
      order 0, for any real number X. The polynomial approximation by
      series of Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
      REFERENCES:
      M.ABRAMOWITZ,I.A.STEGUN, HANDBOOK OF MATHEMATICAL FUNCTIONS, 1965.
      C.W.CLENSHAW, NATIONAL PHYSICAL LABORATORY MATHEMATICAL TABLES,
      VOL.5, 1962.
  --------------------------------------------------------------------- */
static double BessY0 (double X) 
{
		const double P[] = {1.0, 
			-0.1098628627E-2, 
			0.2734510407E-4,
			-0.2073370639E-5, 
			0.2093887211E-6
		};   
		
		const double Q[] = {-0.1562499995E-1, 
			0.1430488765E-3, 
			-0.6911147651E-5,
			0.7621095161E-6, 
			-0.9349451520E-7
		};     
		
		const double R[] = {-2957821389.0, 
			7062834065.0, 
			-512359803.6,
			10879881.29, 
			-86327.92757, 
			228.4622733
		};   
		
		const double S[] = {40076544269.0, 
			745249964.8, 
			7189466.438,
			47447.26470,   
			226.1030244, 
			1.0
		};
      
	return BessY01 (0,
		P, 
		Q,
		R, 
		S,
		X); 
}

/* ---------------------------------------------------------------------
      This subroutine calculates the Second Kind Bessel Function of
      order 1, for any real number X. The polynomial approximation by
      series of Chebyshev polynomials is used for 0<X<8 and 0<8/X<1.
      REFERENCES:
      M.ABRAMOWITZ,I.A.STEGUN, HANDBOOK OF MATHEMATICAL FUNCTIONS, 1965.
      C.W.CLENSHAW, NATIONAL PHYSICAL LABORATORY MATHEMATICAL TABLES,
      VOL.5, 1962.
  ---------------------------------------------------------------------- */
static double BessY1 (double X)
{
	const double P[] = {1.0, 
		0.183105E-2, 
		-0.3516396496E-4,
		0.2457520174E-5, 
		-0.240337019E-6
	};
		
	const double Q[] = {0.04687499995, 
		-0.2002690873E-3, 
		0.8449199096E-5,
		-0.88228987E-6, 
		0.105787412E-6
	};
		
	const double R[] = {-0.4900604943E13, 
		0.1275274390E13, 
		-0.5153438139E11,
		0.7349264551E9,  
		-0.4237922726E7, 
		0.8511937935E4
	};
		
	const double S[] = {0.2499580570E14, 
		0.4244419664E12, 
		0.3733650367E10,
		0.2245904002E8, 
		0.1020426050E6,  
		0.3549632885E3, 
		1.0
	};

	return BessY01 (1,
		P, 
		Q,
		R, 
		S,
		X); 
}

/* -----------------------------------------------------------------
      This subroutine calculates the second kind Bessel Function of
      integer order N, for any real X. We use here the classical
      recursive formula.  
      (2,1) ok, (-2,1) error
  ------------------------------------------------------------------ */
double BessY (int N, double X) 
{
	Access_AMS_Global_Variables;
	
	double TOX, BY, BYM, BYP, R; 
	int J;
	Boolean NEG = N < 0;
	
	N = abs (N);
        
	if (N == 0) { 
		R = BessY0 (X); 
	} else if (N == 1) { 
		R = BessY1 (X); 
	} else if (X == 0.0) { 
		R = FLOATTAB[FPI_NEGINF];
	} else {		
		TOX = 2.0 / X;
		BY  = BessY1 (X);
		BYM = BessY0 (X);
		
		for (J = 1; J < N; J++) {
			BYP = J * TOX * BY - BYM;
			BYM = BY;
			BY  = BYP;
		};
		
		R = BY;
	}      
	
	// when sign of order changes sign of function changes only for odd orders
	return NEG && N % 2 ? -R : R;
}

