/*
		Generic wrapper functions
		
    Copyright (C) 2015  Ralph Schenn

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <tiams.h>
#include "Debug.h"
#include "DSP.h"
#include "Erf.h"
#include "Gamma.h"
#include "MovAvg.h"
#include "ROM.h"
#include "Search.h"
#include "Stack.h"

/* execute function for all arguments 
function to be executed 
simplify the arguments 
requires arguments */
void ExecFuncArgs(void (*funcPtr)(EStackIndex),
	Boolean simp, 
	Boolean reqArgs)
{	
	Access_AMS_Global_Variables; 

	// function requires arguments
	if (reqArgs && remaining_element_count (top_estack) < 1) {
		ER_throwVar (TOO_FEW_ARGS_ERROR);
	}

	// iterate from top to bottom of stack
	// (first argument to last argument)
	while (ESTACK (top_estack) != END_TAG) {
		// function arg should be simplified
		if (simp) {
			push_simplify_statements (top_estack);
		}
		
		// execute func with arg
		(*funcPtr) (top_estack);
		
		// delete internal simplified form of arg
		if (simp) {
			delete_expression (top_estack);
		}
		
		// get next argument
  		top_estack = next_expression_index (top_estack);		
	}
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             	
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* Wrapper for functions with expression and variable 
function to call (returning bool or void) 
called function returns list 
function does not push its result implicitly */
void ExprVar(Boolean (*funcPtr)(EStackIndex, EStackIndex), 
	Boolean list, 
	Boolean push) // 
{	
	Access_AMS_Global_Variables;
	
	EStackIndex expr, var; 
    Boolean b;
    
	// check number of arguments
	CheckNumArgs (2);
	
	// index of unsimplified expression
	expr = top_estack;
	
	// index of unsimplified var
	var = next_expression_index (top_estack);
	
	// simplify the variable
	push_simplify_statements (var);
	
	// simplify the expression
	push_simplify_statements (expr);

	// index of simplified expression
	expr = top_estack;
	
	// index of simplified variable
	var = next_expression_index (top_estack);

	if (list) {
		// function pushes only elments of list to stack
		// but not the enclosing list tags
		push_quantum (END_TAG);
	}

	// function returns bool which must be pushed explicitly
	if (push) {
	    b = (*funcPtr) (expr, var);
		// push result to estack
		push_quantum (bool_to_tag (b));	    
	// function pushes its result to stack automatically
	} else {
		(*funcPtr) (expr, var);
	}
	
	if (list) {
		// function pushes only elments of list to stack
		// but not the enclosing list tags
		push_quantum (LIST_TAG);
	}

	// remove the simplified args from stack
	delete_between (next_expression_index (var), expr);
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif               
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* call func with one expr parameter */
void IsExpr(Boolean (*funcPtr)(EStackIndex i)) // 
{
	Access_AMS_Global_Variables; 
	Boolean b;
	
	// check number of arguments
	CheckNumArgs (1);
	
	// simplify to internal form
	push_simplify_statements (top_estack);
	
	// call is function
	b = (*funcPtr) (top_estack);
	
	// delete the internal simplified function arg
	delete_expression (top_estack);
				
	// push result to estack
	push_quantum (bool_to_tag (b)); 
	
	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}   

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define LEN 1

/* function receiving list and integer */
void LstIntFunc(void (*funcPtr)(double [], unsigned short, unsigned short, double [])) // 
{       
    Access_AMS_Global_Variables;

    volatile HANDLE xHdl, yHdl;  
	
	double *xPtr, *yPtr;

    unsigned short lenX, lenY = 0;
    short lenOut;             

	// check number of parameters
	CheckNumArgs (2);
                
	// get second argument 
	lenOut = GetValue (EX_getArg (LEN), 1, MAX_LST_LEN);
	        
	TRY	
		// read list of into array (in reverse order)
		// (automatically convert to float if necessary)
		lenX = GetFloatListArg (top_estack, &xHdl);
		
		// input list is not empty and 
		// length of average not longer than length of input list 
		if ((funcPtr == &MovAvg && lenX > 0 && lenOut <= lenX) ||
			funcPtr != &MovAvg) {
			// calculate length of output list
			if (funcPtr == &MovAvg) {
				lenY = lenX - lenOut + 1;
			} else if (funcPtr == &Downsamp) {
				// cast to double before division
				// otherwise we are dividing integers and get integer result
				// which is to low
				lenY = ceil ((double)lenX / (double)lenOut);
			} else if (funcPtr == &Upsamp) {
				lenY = lenX * lenOut;
			}
							
			// allocate output list
			yHdl = HeapAllocThrow (lenY * sizeof(double));
			
			// dereference after the call to HeapAlloc 
			// and also after the call to HeapRealloc (called by GetFloatListArg)
			xPtr = HeapDeref (xHdl);
			yPtr = HeapDeref (yHdl);
				
			// calculate the moving average
			(*funcPtr)(xPtr,
				lenX, 
				lenOut,
				yPtr); 
			
			// push list to estack (in reverse order)
			push_float_list (yPtr, lenY);
		} else {
			ER_throwVar (ER_DOMAIN);		
		}
	FINALLY
		// HeapFreeIndir handles H_NULL correctly
		HeapFreeIndir ((HANDLE *)&xHdl);
		HeapFreeIndir ((HANDLE *)&yHdl);
	ENDFINAL
}                    

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* intersection or difference of lists
map_tail is not suitable here because callback function also needs pointer to other list */
void LstIntSub(Boolean sub)
{
	Access_AMS_Global_Variables;
	
	EStackIndex lst1 = top_estack, 
		lst2,
		lst1s,
		lst2s,
		lst1so;
	
	CheckNumArgs (2);
	
	lst2 = next_expression_index (top_estack);
	
	push_simplify_statements (lst2);
	push_simplify_statements (lst1);
		
	lst1s = top_estack;
	lst2s = next_expression_index (top_estack);
	
	if (!is_list (lst1s) || !is_list (lst2s)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	push_quantum (END_TAG);
	
	// save index for later deletion of list
	lst1so = lst1s;
	
	// start with first element of lst1
	lst1s--;
	
	// loop over first list
	while (ESTACK (lst1s) != END_TAG) {
		// difference and element found in second list or
		if (!sub && BinSearch (lst2s, lst1s) > 0 ||
			// subtraction and element not found in second list
			sub && BinSearch (lst2s, lst1s) == 0) {
			// add element of first list to result list
			push_expression (lst1s);
		}
		
		// next element of first list
		lst1s = next_expression_index (lst1s);
	}

	// terminate the difference
	push_quantum (LIST_TAG); 
	
	// remove simplified lists
	delete_between (lst2s, lst1so); 
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* odd or even elements of list */
void LstOddEven(Boolean odd)
{   
	Access_AMS_Global_Variables;
	
	volatile HANDLE hdl;
	EStackIndex *ptr;
	unsigned short len;
	short i;
	
	CheckNumArgs (1);
    
    TRY
		len = GetExpressionListArg (top_estack, &hdl); 
				
		ptr = HeapDeref (hdl);
		
		push_quantum (END_TAG);
        
        // calculate right index at which we start 
        i = len - (odd ? ((len % 2) ? 1 : 2) : ((len % 2) ? 2 : 1));  
        
		// loop from right to left (avoids result in reversed order)
		// decrementing by two saves check if index is even
		for ( ; i >= 0; i -= 2) {
			push_expression (ptr[i]);
		}
		
		push_quantum (LIST_TAG);
    FINALLY
		HeapFreeIndir ((HANDLE *)&hdl);    
    ENDFINAL
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* proccess all arguments of list with passed function */ 
void MapTail(void (*funcPtr) (EStackIndex))
{
	Access_AMS_Global_Variables;
	
	EStackIndex argptr, old_top;
	
	CheckNumArgs (1);
	
	push_simplify_statements (top_estack);
	
	if (!is_list (top_estack)) {
		ER_throwVar (ER_ARGUMENT_MUST_BE_LIST);
	}
	
	old_top = top_estack;
	
	// first element of list
	argptr = top_estack - 1;
		
	map_tail (funcPtr, argptr);	

	push_quantum (LIST_TAG);
	
	// remove simplified list
	delete_expression (old_top);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* matrix product 
not much faster than builtin function 
but a good starting point for functions requiring matrices as args */
/*void matprod(void)
{
    Access_AMS_Global_Variables;  

	EStackIndex argptr = top_estack;
	volatile HANDLE aHdl = H_NULL;
	volatile HANDLE bHdl = H_NULL;
	volatile HANDLE abHdl = H_NULL;
	
	double *aPtr;
	double *bPtr;
	double *abPtr;
	
	unsigned short rowsA = 0, colsA = 0;
	unsigned short rowsB = 0, colsB = 0;

	// check number of parameters
	CheckNumArgs (2);		

	TRY		 
		// read matrix A
		GetFloatMatArg (argptr, 
			&aHdl,
			&rowsA,
			&colsA);
			
		// get index of matrix B
		argptr = next_expression_index (argptr);
				
		// read matrix B
		GetFloatMatArg (argptr, 
			&bHdl,
			&rowsB,
			&colsB);
			
		if (colsA == rowsB) {
			// alloacte matrix AB
			abHdl = HeapAllocThrow (rowsA * colsB * sizeof (double));
	
			// matrix A and B have been read succesfully
			if (aHdl != H_NULL && bHdl != H_NULL && abHdl != H_NULL)
			{
				aPtr = HeapDeref (aHdl);
				bPtr = HeapDeref (bHdl);
				abPtr = HeapDeref (abHdl);
				
				// calculate the matrix product
				MatProd(aPtr, 
					rowsA, 
					colsA, 
					bPtr, 
					colsB,
					abPtr);

				// push matrix product to estack
				push_float_mat (abPtr, rowsA, colsB);
		  	}	       
	  	} else {
	  		ER_throwVar (ER_DIMENSION);
	  	}
	FINALLY
		// deallocate the matrices
		HeapFreeIndir ((HANDLE *)&aHdl);
		HeapFreeIndir ((HANDLE *)&bHdl);
		HeapFreeIndir ((HANDLE *)&abHdl);
	ENDFINAL
}*/

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* wrapper for functions requiring real arguments
handles matrices, lists or scalars */ 
void RealFunc(void (*funcPtr)(EStackIndex))
{
	Access_AMS_Global_Variables;

	EStackIndex argptr, before_rev, simp; 
    
	// check number of parameters
	CheckNumArgs (1);
    
	push_simplify_statements (top_estack);
	
	// to remove simplified argument after pushing result
	simp = top_estack;
	
	// argument is matrix
	if (is_matrix (top_estack)) {
		argptr = top_estack; 
		
		// end of matrix
		push_quantum (END_TAG);
		
		// begin of first row
		argptr--;
		
		// loop over rows
		while (ESTACK (argptr) != END_TAG) {
			// apply function to the row
			map_tail (funcPtr, argptr - 1);
			
			// push the LIST_TAG on top of the row
			push_quantum (LIST_TAG);
			
			// point to next row
			argptr = next_expression_index (argptr); 		
		}		

		// begin of matrix
		push_quantum (LIST_TAG);
		
		// save to remove later
		before_rev = top_estack;
		
		// reverse the order of the rows of the result
		push_reversed_tail (top_estack - 1); 
			
		// beginning of reversed matrix
		push_quantum (LIST_TAG);
		
		// remove matrix before row reversal
		delete_expression (before_rev);	
	// argument is list
	} else if (ESTACK (top_estack) == LIST_TAG) {
		// apply function to the tail
		map_tail (funcPtr, top_estack - 1);
		
		// push the LIST_TAG on top of the tail
		push_quantum (LIST_TAG);
	// argument is scalar
	} else {
		(*funcPtr) (top_estack);	
	}
		
	// remove simplified argument
	delete_expression (simp);

	#ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* string function with one arg wrapper */
void StrFunc(const char * (*funcPtr)(const char *))
{
	Access_AMS_Global_Variables;    
    const char *str;
    
	CheckNumArgs (1);
	
    push_simplify_statements (top_estack);
    
    str = GetStrnArg (top_estack);
    
	(*funcPtr) ((char *)str);
	
	// delete after using the pointer to str on estack
    delete_expression (top_estack);
    
    #ifdef DEBUG
		PushStackAsMat ();
	#endif             
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/* set arbitary number */
void SetArb(Boolean integer)
{
	Access_AMS_Global_Variables;
	 
	unsigned short a;     
	
	CheckNumArgs (1);
	
	// negative or > 255
	a = estack_to_ushort_simp (top_estack);
	
	if (a > UCHAR_MAX) {
		ER_throwVar (ER_DOMAIN);
	}
	
    if (integer) {
		ARb_int_count = a;
	} else {
		ARb_real_count = a;	
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*void Lst2Scal(void)
{
    Access_AMS_Global_Variables;

    volatile HANDLE xHdl = H_NULL;
	
	double *xPtr;

    unsigned short lenX;

	// check number of parameters
	CheckNumArgs (1);
                
	TRY	
		// read list of into array (in reverse order)
		// (automatically convert to float if necessary)
		lenX = GetFloatListArg (top_estack, &xHdl);
		
		// input list is not empty and 
		// length of average not longer than length of input list
		if (lenX > 0) {
			// calculate length of output list
			// dereference after the call to HeapAlloc 
			// and also after the call to HeapRealloc (called by GetFloatListArg)
			xPtr = HeapDeref (xHdl);
				
			// push list to estack (in reverse order)
			push_Float (RMS (xPtr, lenX));
		} else {
			ER_throwVar (ER_DOMAIN);		
		}
	FINALLY
		// HeapFreeIndir handles H_NULL correctly
		HeapFreeIndir ((HANDLE *)&xHdl);
	ENDFINAL
}*/
